/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 07/04/2003 */

#include "clustergraph.h"
#include "../readwrite/assume.h"
#include "../readwrite/populate.h"
#include "findhvclusters.h"

clustergraph::clustergraph(cluster& cl, unsigned nc):
  hvclmem(cl.getbinsize(),sizeof(clgnode)+sizeof(onedimlinks)*
	  (cl.getnumthings()-1)+sizeof(sint8)*nc),
  counteroffset(sizeof(clgnode)+sizeof(onedimlinks)*(cl.getnumthings()-1)){
  c=&cl;
  numcounters=nc;
  root=NULL;
  traversalorder=NULL;
}

clustergraph::~clustergraph(){
  if(traversalorder!=NULL){
    for(int i=hvclmem.size()-1;i>=0;i--){
      for(int j=0;j<c->getnumthings();j++){
	freelist(traversalorder[i]->links[j].children);
      }
    }
    free(traversalorder);
  }
}

void clustergraph::read_in_clusters(char* rawclusterfile, unsigned cnum){
  assume(cnum<numcounters,"Why ask poor clustergraph to use inexistent counter?\n");
  read_plain(rawclusterfile, *c, hvclmem, AGGREGATION_SUM,
	     counteroffset+sizeof(sint8)*cnum);
  /* Fix pointers and add ancestors of clusters from the file that
     were't there. */
  add_missing_pointers_and_nodes();
  rebuild_traversal_order();
  /* Set the traffic of these new nodes based on the traffic of their
     children */
  for(int i=hvclmem.size()-1;i>=0;i--){
    for(int j=0;j<c->getnumthings();j++){
      sint8 childsum=sumall(traversalorder[i]->links[j].children,cnum);
      if (childsum>counter_i(traversalorder[i],cnum)){
	counter_i(traversalorder[i],cnum)=childsum;
      }
    }
  }
}

void clustergraph::find_hvclusters(record& flowID, hash_map& flow_memory,
				   sint8 threshold, unsigned cnum){
  assume(cnum<numcounters,"Why ask poor clustergraph to use inexistent counter?\n");
  find_high_volume_clusters(flowID,flow_memory,threshold,hvclmem,
			    counteroffset+sizeof(uint8)*cnum);
  add_missing_pointers_and_nodes();
  rebuild_traversal_order();
}

/* adds to a list ordered by the value of the pattern number pnum */
clgnlist* clustergraph::addtoorderedlist(clgnlist* l, clgnode* n, 
					 unsigned pnum){
  if(l==NULL || c->getpattern(pnum).comesfirst(
		(byte*)n-c->getbinsize()+c->getoffset(pnum),
		(byte*)(l->node)-c->getbinsize()+c->getoffset(pnum))){
    clgnlist* result=(clgnlist*)malloc(sizeof(clgnlist));
    assumenotnull(result);
    result->node=n;
    result->next=l;
    return result;
  }
  l->next=addtoorderedlist(l->next,n,pnum);
  return l;
}

/* This function makes sure that all the parent/child pointers are
   right. In the proces it might need to create certain new nodes
   coresponding to clusters whose children were there, but they
   weren't. */
#define node ((clgnode*)j.value())
void clustergraph::add_missing_pointers_and_nodes(){
  byte parent[300];
  assume(c->getbinsize()<=300,"Oops, cluster size larger than 300\n");
  for(growing_hash_iterator j(hvclmem);!j.at_end();++j){
    for(int i=0;i<c->getnumthings();i++){
      if(!c->getpattern(i).isstar(j.key()+c->getoffset(i))){
	if(node->links[i].parent==NULL){ /* Must add parent pointer */
	  copy(j.key(),parent,c->getbinsize());
	  c->getpattern(i).settoparent(j.key()+c->getoffset(i),
				       parent+c->getoffset(i));
	  clgnode* pnode=((clgnode*)hvclmem[parent]);
	  node->links[i].parent=pnode;
	  pnode->links[i].children=addtoorderedlist(pnode->links[i].children,
						    node,i);
	}
      }
    }
  }
}
#undef node

sint8 clustergraph::sumall(clgnlist* nodes,unsigned cnum){
  sint8 result=0;
  while(nodes!=NULL){
    result+=counter_i(nodes->node,cnum);
    nodes=nodes->next;
  }
  return result;
}


void clustergraph::rebuild_traversal_order(){
  if(traversalorder!=NULL){
    free(traversalorder);
  }
  if(hvclmem.size()==0){
    traversalorder=NULL;
    root=NULL;
    return;
  }
  traversalorder=(clgnode**)malloc(sizeof(clgnode*)*hvclmem.size());
  assumenotnull(traversalorder);
  int i=0;
  for(growing_hash_iterator j(hvclmem);!j.at_end();++j){
    traversalorder[i]=(clgnode*)j.value();
    i++;
  }
  assume(c->getbinsize()<=300,"Oops, cluster size larger than 300\n");
  sortclusters(0,hvclmem.size()-1);
  root=traversalorder[0];
}


static byte pivot[300];
#define key(i) ((byte*)traversalorder[i]-c->getbinsize())
void clustergraph::sortclusters(int first, int last) {
  int sizeminus1 = last - first;
  unsigned int i,j,lastbeforepivot;
  if (sizeminus1 <1) {
    return;
  }
  copy(key(first+sizeminus1/2),pivot,c->getbinsize());
  i=j=first;
  while(j<=last){
    if (c->comesfirst(key(j),pivot)){
      if (i != j) {
	clgnode* a=traversalorder[i];
	traversalorder[i]=traversalorder[j];
	traversalorder[j]=a;
      }
      i++;
    }
    j++;
  }
  lastbeforepivot=i-1;
  j=i;
  while(j<=last) {
    if(equals(key(j),pivot,c->getbinsize())){
      if (j !=i ){
	clgnode* a=traversalorder[i];
	traversalorder[i]=traversalorder[j];
	traversalorder[j]=a;
      }
      i++;
    }
    j++;
  }
  sortclusters(first,lastbeforepivot);
  sortclusters(i,last);
}
#undef key

void freelist(clgnlist*l){
  clgnlist* crt=l; 
  while(crt!=NULL){
    l=crt->next;
    free(crt);
    crt=l;
  }
}

clgnlist* addtolist(clgnlist* l, clgnode* n){
  clgnlist* result=(clgnlist*)malloc(sizeof(clgnlist));
  assumenotnull(result);
  result->node=n;
  result->next=l;
  return result;
}

int listlength(clgnlist*l){
  int result=0;
  while(l!=NULL){
    l=l->next;
    result++;
  }
  return result;
}
