/* See AutoFocus/doc/COPYING for copyright/licensing information. */
#include <stdlib.h>
#include <assert.h>
#include "delta.h"

void normalizeclusterreports(clustergraph*cg){
  clgnode** to=cg->traversalorder;
  unsigned traffcoff=cg->counteroffset;
  unsigned expectedcoff=cg->counteroffset+sizeof(sint8);
  double ratio=((double)counteri(cg->root,traffcoff))/
    (double)counteri(cg->root,expectedcoff);
  for(int i=cg->hvclmem.size()-1;i>=0;i--){
    counteri(to[i],expectedcoff)=(sint8)(counteri(to[i],expectedcoff)*ratio);
  }
}

static unsigned traffcoff;
static unsigned expectedcoff;
static sint8 threshold;
static cluster* c;
inline const byte* key(clgnode* node){return (byte*)node-c->getbinsize();}
static sint8 maxexplanation;
/* The largest explanation offered by a maximal subset of the reported
   clusters for the current cluster. */
static sint8 minexplanation;
/* The smallest explanation offered by a maximal subset of the
   reported clusters for the current cluster. */
 
sint8 add_up_deltas(clgnlist* l){
  sint8 result=0;
  sint8 diff;
  for(;l!=NULL;l=l->next){
    diff=counteri(l->node,traffcoff)-counteri(l->node,expectedcoff);
    result+=diff;
  }
  return result;
}

/* This function tests whether the cluster defined by cl intersects
   any of the clusters in the list. */
int does_intersect_any(const byte* cl, clgnlist* clusters){
  for(;clusters!=NULL;clusters=clusters->next){
    switch(c->compare(cl,key(clusters->node))){
    case C_SECOND_INCLUDES_FIRST:
    case C_FIRST_INCLUDES_SECOND:
    case C_OVERLAPPING:
      return 1;
      break;
    case C_NON_OVERLAPPING:
      break;
    case C_EQUAL:
      abort(); // The cluster should not be in the list
    default:
      abort(); // We handled all posssible return values
    }    
  }
  return 0;
}

/* This function tests whether the cluster defined by cl can replace
   one or more clusters in the list (by being more general) without
   overlapping with any of the others. */
int can_replace_some(const byte* cl, clgnlist* clusters){
  int result=0;
  for(;clusters!=NULL;clusters=clusters->next){
    switch(c->compare(cl,key(clusters->node))){
    case C_FIRST_INCLUDES_SECOND:
      result=1;
      break;
    case C_OVERLAPPING:
    case C_SECOND_INCLUDES_FIRST:
      return 0;
    case C_NON_OVERLAPPING:
      break;
    case C_EQUAL:
      abort(); // The cluster should not be in the list
    default:
      abort(); // We handled all posssible return values
    }    
  }
  return result;
}

/* Given the clusters in the knownclusters list, this function builds
   all possible maximal subsets of nonoverlapping clusters. The subset
   list contains the clusters in the subset and the notinsubset
   contains those not it int. The knownclusters are those that still
   have to be assigned to one of the two sets. Nonoverlapping subset
   means noe of its clusters overlap. Maximal subset means more
   specific in what it covers than another nonoverlapping subset. This
   function recursively goes through the maximal nonoverlapping
   subsets and it stores into maxexplanation and minexplanation the
   largest and smallest sum of elements within a subset found so far. */
static void find_maximal_explanations(clgnlist* knownclusters,
				      clgnlist* subset,
				      clgnlist* notinsubset){
  if(knownclusters==NULL){/* We finished building the subset */
    if(subset==NULL){
      return; /* This literally tells us nothing. */
    }
    /* Check if we left out any clusters that don't overlap with any
       of those in the subset. (If there is one, than adding that
       cluster to the subset would give us a larger subset that
       contains this one, so we should not use this one for any
       predictions.) */
    for(;notinsubset!=NULL;notinsubset=notinsubset->next){
      if(!does_intersect_any(key(notinsubset->node),subset)){
	/* This is not a maximal subset. */
	return;
      }
      /* If it's more general than some without overlapping with the
	 others the subset is not maximal. */
      if(can_replace_some(key(notinsubset->node),subset)){
	return;
      }
    }
    sint8 explanation=add_up_deltas(subset);
    if(explanation>maxexplanation){
      maxexplanation=explanation;
    }
    if(explanation<minexplanation){
      minexplanation=explanation;
    }
    return;
  } else { /* knownclusters!=NULL */
    clgnlist* newknown=knownclusters->next;
    if(does_intersect_any(key(knownclusters->node),subset)){
      /* We cannot add it to the subset. */
    } else {
      /* Let's try to add it to the subset */
      knownclusters->next=subset;
      find_maximal_explanations(newknown,knownclusters,notinsubset);
    }
    /* Let's try to leave it out. */
    knownclusters->next=notinsubset;
    find_maximal_explanations(newknown,subset,knownclusters);
    /* Let's restore things. */
    knownclusters->next=newknown;
  }
}

/* If all the clusters known so far (this includes all more specific
   clusters that will ever make it into the report) give estimates for
   our cluster within the threshold of its actual size, this function
   returns true, otherwise our cluster is supposed to be put into the
   compressed delta report. */
static int explains_unambiguously(clgnlist* reportedclusters,
				  clgnode* ourcluster){
  clgnlist* msc=NULL; // The reported clusters that are more specific than ours
  for(;reportedclusters!=NULL;reportedclusters=reportedclusters->next){
    switch(c->compare(key(reportedclusters->node),key(ourcluster))){
    case C_SECOND_INCLUDES_FIRST:
      msc=addtolist(msc,reportedclusters->node);
      break;
    case C_NON_OVERLAPPING:
    case C_OVERLAPPING:
      break;
    case C_EQUAL:
    case C_FIRST_INCLUDES_SECOND:
      abort(); // The reportedclusters "does not match" our cluster
    default:
      abort(); // We handled all posssible return values
    }
  }
  sint8 diff=counteri(ourcluster,traffcoff)-counteri(ourcluster,expectedcoff);
  if(msc!=NULL){
    if(listlength(msc)>60){
      // The algorithm would probably take forever
      buffer b;
      c->print(b,key(ourcluster));
      b.print_char('\0');
      fprintf(stderr,"Too many more specific clusters (%d) for (%s) skipping delta compression.\n",listlength(msc),b.buff);
      return 0;
    }
    maxexplanation=diff-threshold-1; // same as -infinity
    minexplanation=diff+threshold+1; // same as +infinity
    find_maximal_explanations(msc,NULL,NULL);
    assert((maxexplanation!=diff-threshold-1) ||
	   (minexplanation!=diff+threshold+1));
    freelist(msc);
    if(maxexplanation>=diff+(long long)threshold || 
       minexplanation<=diff-(long long)threshold){
      return 0;
    } else {
      return 1;
    }
  } else {
    if(diff>=threshold || -diff>=threshold){
      return 0;
    } else {
      return 1;
    }
  }
}

clgnlist* build_delta_report(clustergraph* cg, sint8 th){
  clgnode** to=cg->traversalorder;
  traffcoff=cg->counteroffset;
  expectedcoff=cg->counteroffset+sizeof(sint8);
  threshold=th;
  clgnlist* result=NULL;
  c=cg->c;
  for(int i=cg->hvclmem.size()-1;i>=0;i--){
    if(!explains_unambiguously(result,to[i])){
      result=addtolist(result,to[i]);
    }
  }
  return result;
}
