/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 06/28/2002 */

#include "findhvclusters.h"
#include "../readwrite/assume.h"
#include "onefield/ofh.h"
#include "onefield/ofm.h"
#include "../hash_map/growing_hash.h"

static growing_hash* smallclusters;
int isknowntobesmall(const byte* cluster); 
/* checks whether any of the parents of the given cluster are small*/
#define SMALLCLUSTERMEGS   64
/* How much memory we can afford for the small cluster growing_hash. As a
   rule of thumb this should be one quarter of the memory of the
   computer running autofocus. */

int maybeabovethreshold(const byte* cluster);
/* goes through the flows to see whether a cluster is large. Because
   it batches, it cannot always tell immediately whether a cluster is
   large. This function populates the hvclusters hash_map as a side
   effect. */
#define BATCHSIZE 100
byte* clbatchmemory;
int clbatchcounter;
int flushclbatch(); /* computes sizes for the clusters in the batch */

void clean_smallclusters();
/* To reduce memory usage we clean the smallclusters hash_map when we
   advance the first field. */

/* The arguments of find_high_volume_clusters are made available to
   the other functions */
static record* ID;
static hash_map* hvclusters;
static hash_map* flows;
static uint8 th;
static unsigned offset;
static ofhorder* orders[MAX_COMPOUND_SIZE];
static ofhiterator* iterators[MAX_COMPOUND_SIZE];
static int foundlarge[MAX_COMPOUND_SIZE];
void find_high_volume_clusters(record& flowID, hash_map& flow_memory,
			       uint8 threshold, hash_map& result,
			       unsigned offst){
  hvclusters=&result;
  ID=&flowID;
  flows=&flow_memory;
  th=threshold;
  offset=offst;
  int nfields=flowID.getnumthings();
  /* We first build the traversal orders for large clusters along all
     dimensions */
  for(int i=0;i<nfields;i++){
    ofm* crtofm=flowID.getcluster().getpattern(i).buildofm();
    hash_map_iterator* j=flow_memory.build_iterator();
    for(;!j->at_end();++(*j)){
      crtofm->newpacket(j->key()+flowID.getoffset(i),*(uint8*)j->value());
    }
    delete j;
    orders[i]=crtofm->buildorderoflargepatterns(threshold);
    delete crtofm;
  }
  smallclusters=new growing_hash(flowID.getcluster().getbinsize(),0);
  assumenotnull(smallclusters);
  for(int i=0;i<nfields;i++){
    iterators[i]=new ofhiterator(*orders[i]);
    assumenotnull(iterators[i]);
  }
  clbatchmemory=(byte*)malloc(BATCHSIZE*flowID.getcluster().getbinsize());
  assumenotnull(clbatchmemory);
  clbatchcounter=0;
  byte* crtcluster=(byte*)malloc(flowID.getcluster().getbinsize());
  assumenotnull(crtcluster);
  flowID.getcluster().setwildcards(crtcluster);
  for(int i=0;i<nfields;i++){foundlarge[i]=0;}
  while(iterators[0]->havepattern()){
    int i;
    if(!isknowntobesmall(crtcluster)){
      if(maybeabovethreshold(crtcluster)){
	for(i=nfields-1;i>=0 && foundlarge[i]==0;i--){
	  foundlarge[i]=1;
	}
      }
    }
    /* Moving on to the next cluster */
    iterators[nfields-1]->next(foundlarge[nfields-1]);
    foundlarge[nfields-1]=0;
    for(i=nfields-1;i>0 && !iterators[i]->havepattern();i--){
      iterators[i-1]->next(foundlarge[i-1]);
      foundlarge[i-1]=0;
    }
    if(i==0){
      if(!iterators[0]->havepattern()){
	continue; /* End of the while loop */
      }
      clean_smallclusters();
    }
    copy(iterators[i]->getpattern(),
	 crtcluster+flowID.getcluster().getoffset(i),
	 flowID.getcluster().getpattern(i).getbinsize());
    for(i=i+1;i<nfields;i++){
      iterators[i]->reset();
      copy(iterators[i]->getpattern(),
	   crtcluster+flowID.getcluster().getoffset(i),
	   flowID.getcluster().getpattern(i).getbinsize());
    }
  }
  flushclbatch();
  for(int i=0;i<nfields;i++){
    delete iterators[i];
    delete orders[i];
  }
  free(clbatchmemory);
  delete smallclusters;
}

byte parentcluster[200];

int isknowntobesmall(const byte* cluster){
  if(smallclusters->has(cluster)){
    return 1; /* This should never happen */
  }
  assume(ID->getcluster().getbinsize()<=200,
	 "Fixed cluster buffer size too small \n");
  copy(cluster,parentcluster,ID->getcluster().getbinsize());
  for(int i=0;i<ID->getnumthings();i++){
    if(!ID->getcluster().getpattern(i).isstar(iterators[i]->getpattern())){
      copy(iterators[i]->getparentpattern(),
	   parentcluster+ID->getcluster().getoffset(i),
	   ID->getcluster().getpattern(i).getbinsize());
      if(smallclusters->has(parentcluster)){
	smallclusters->insert(cluster);
	return 1;
      } else { /* restore parentcluster to cluster */
	copy(iterators[i]->getpattern(),
	     parentcluster+ID->getcluster().getoffset(i),
	     ID->getcluster().getpattern(i).getbinsize());
      }
    }
  }
  return 0;
}

int maybeabovethreshold(const byte* cluster){
  copy(cluster,clbatchmemory+ID->getcluster().getbinsize()*clbatchcounter,
       ID->getcluster().getbinsize());
  clbatchcounter++;
  if(clbatchcounter<BATCHSIZE){
    return 1;
  } else {
    return flushclbatch();
  }
}

int flushclbatch(){ /* reurns true if last cluster is large */
  if(clbatchcounter==0){
    return 0;
  }
  uint8 clustertraffic[BATCHSIZE];
  for(int i=0;i<clbatchcounter;i++){clustertraffic[i]=0;}
  hash_map_iterator* j=flows->build_iterator();
  for(;!j->at_end();++(*j)){
    for(int i=0;i<clbatchcounter;i++){
      if(ID->getcluster().matches(clbatchmemory+ID->getcluster().getbinsize()*i,j->key())){
	clustertraffic[i]+=*(uint8*)j->value();
      }
    }
  }
  delete j;
  for(int i=0;i<clbatchcounter;i++){
    if(clustertraffic[i]<th){
      smallclusters->insert(clbatchmemory+ID->getcluster().getbinsize()*i);
    } else {
      *(uint8*)((*hvclusters)[clbatchmemory+ID->getcluster().getbinsize()*i]
		+offset)=clustertraffic[i];
    }
  }
  if(clustertraffic[clbatchcounter-1]<th){
    clbatchcounter=0;
    return 0;
  } else {
    clbatchcounter=0;
    return 1;
  }
}

void clean_smallclusters(){
  if(smallclusters->size()>SMALLCLUSTERMEGS*1024*1024/
     (2*sizeof(entrylist*)+ID->getcluster().getbinsize())){
    delete smallclusters;
    smallclusters=new growing_hash(ID->getcluster().getbinsize(),0);
  }
}
