/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 06/24/2003 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "ipofm.h"
#include "prefix.h"
#include "../../readwrite/assume.h"

ipofm::ipofm():counters(4,8){
}

void ipofm::newpacket(const byte* fvalue, uint8 increment){
  (*(uint8*)(counters[fvalue]))+=increment;
}

ofhorder* ipofm::buildorderoflargepatterns(uint8 threshold){
  ipofmorder* result =new ipofmorder(counters, threshold);
  assumenotnull(result);
  return result;
}

static uint4* addresses;
static void sortaddresses(int first, int last);
/* Sorts into numerically ascending order the addresses from the array
   pointed to by "addresses" starting from position first and up to
   position last. */

static void compute_traversal_order(hash_map& counters, hash_map& result, 
				    uint8 threshold);
/* The function that computes hash_map with the traversal order */

ipofmorder::ipofmorder(growing_hash& counters, uint8 threshold){
  uint8 totaltraffic=0;
  for(growing_hash_iterator j(counters);!j.at_end();++j){
    totaltraffic+=*(uint8*)(j.value());
  }
  if(totaltraffic < threshold){
    numpatterns=0;
    patterns=NULL;
    return;
  }
  /* First we sort the IP addresses. This gives us the order we would
     see them in if we traversed the IP hierarchy tree. */
  addresses=(uint4*)malloc(4*counters.size());
  assumenotnull(addresses);
  int i=0;
  for(growing_hash_iterator j(counters);!j.at_end();++j){
    addresses[i]=*((uint4*)(j.key()));
    i++;
  }
  assert(i==counters.size());
  sortaddresses(0,i-1);

  /* Next, using the sorted IP addresses, we compute the traversal
     order and store it into the largeprefixes growing_hash. */
  growing_hash largeprefixes(5,sizeof(pmh_entry));
  compute_traversal_order(counters, largeprefixes,threshold);
  free(addresses);

  /* Finally we build the actual order array. */
  convert_growing_hash(largeprefixes);
}

static void compute_traversal_order(hash_map& counters, hash_map& result, 
				    uint8 threshold){
/* We traverse the hierarchy by visiting the leaves in order. At all   times we store in the prefixes array the 25 ancestors of the
   current leaf and the leaf. As we visit all the leaves, this array
   "sweeps" the part of the IP prefix hierarchy covered by the
   traffic. We use this array to generate the inorder traversal of the
   high volume prefixes. */
  struct paentry{
    prefixstruct prefix;
    int index; /* The position of this prefix in the traversal. */
    uint8 traffic;
    phm_entry* entry; /* Valid iff index!=-1 */ 
  } pa[26];
  int depth;
  for(depth=1;depth<26;depth++){
    pa[depth].index=-1;
    pa[depth].prefix.ipaddr=0;
    pa[depth].prefix.masklen=depth+7;
    pa[depth].traffic=0;
  }
  pa[0].index=-1;
  pa[0].prefix.ipaddr=0;
  pa[0].prefix.masklen=0;
  int crtindex=0;
  /* When an entry in the array reaches the threshold it gets an index
     number and gets added to the hash_map. The one before it in the
     array will be its father. The pruneindex of a prefix is known
     only when we remove it from the array because it no longer
     matches the current IP address. */
  for(int j=0;j<counters.size();j++){
    uint4 crtip=addresses[j];
    uint8 crttraffic=*(uint8*)(counters[(byte*)(&crtip)]);
    /* Remove prefixes that don't match */
    for(depth=25;prefixoflengthx(crtip,pa[depth].prefix.masklen)!=
	  pa[depth].prefix.ipaddr;depth--){
      if(pa[depth].index!=-1){
	pa[depth].entry->pruneindex=crtindex;
      }
      pa[depth].prefix.ipaddr=prefixoflengthx(crtip,pa[depth].prefix.masklen);
      pa[depth].index=-1;
      pa[depth].traffic=0;
    }
    /* Update the prefixes that match and are not in result already */
    for(depth=25;depth>=0 && pa[depth].index==-1;depth--){
      pa[depth].traffic+=crttraffic;
    }
    /* Add to the result those that rose above the threshold */
    for(depth=depth+1;depth<26 && pa[depth].traffic>=threshold;depth++){
      pa[depth].index=crtindex;
      pa[depth].entry=(pmh_entry*)result[(byte*)(&(pa[depth].prefix))];
      if(depth==0){
	pa[depth].entry->parentindex=0;
      } else {
	pa[depth].entry->parentindex=pa[depth-1].index;
      }
      crtindex++;
    }
  }
  for(depth=25;depth>=0;depth--){
    if(pa[depth].index!=-1){
      pa[depth].entry->pruneindex=crtindex;
    }
  }
}

inline void swap(unsigned int i,unsigned int j) {
  uint4 a=addresses[i];
  addresses[i]=addresses[j];
  addresses[j]=a;
}

static void sortaddresses(int first, int last) {
  int sizeminus1 = last - first;
  uint4 pivot;
  unsigned int i,j;
  if (sizeminus1 <1) {
    return;
  }
  pivot = addresses[first+sizeminus1/2];
  i=j=first;
  while(j<=last){
    if (addresses[j] < pivot){
      if (i != j) {
	swap(i,j);
      }
      i++;
    }
    j++;
  }
  sortaddresses(first,i-1);
  j=i;
  while(j<=last) {
    if(addresses[j] == pivot){
      if (j !=i ){
	swap(i,j);
      }
      i++;
    }
    j++;
  }
  sortaddresses(i,last);
}

