/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 07/13/2003 */

/* This is the main program that preforms three pieces of core
   functionality: computing uncompressed reports, compressed reports
   and compressed delta reports. The input for uncompressed reports is
   a traffic trace. The input for compressed reports can be a traffic
   trace, an uncompressed report or a compressed report. The input for
   a compressed delta report are two compressed or uncompressed
   reports. */

#include <stdlib.h>
#include <stdio.h>
#include "../getopt/getopt.h"
#include "../readwrite/assume.h"
#include "../datatypes/ip/ip.h"
#include "../datatypes/port/port.h"
#include "../datatypes/protocol/proto.h"
#include "../datatypes/cluster.h"
#include "../algorithms/clustergraph.h"
#include "../algorithms/delta.h"
#include "../algorithms/compression.h"
#include "../algorithms/findhvclusters.h"
#include "../algorithms/countingsynopsis.h"
#include "../algorithms/flowsynopsis.h"
#include "../readwrite/populate.h"

static struct option long_options[] =
{
  {"trace",1,0,0}, // the trace
  {"flowrecfile",1,0,0}, // the input file with flow records
  {"clusters",1,0,0}, // the cluster report describing the traffic
  {"expected",1,0,0}, // the cluster report we compare against (delta)
  {"resolution",1,0,0}, // threshold = totaltraffic / resolution 
  {"uncompressed",0,0,0}, // do not compress traffic report
  {"normalize",0,0,0}, // normalize the expected traffic (delta)
  {"countpackets",0,0,0}, // count packets instead of bytes
  {"countflows",0,0,0}, // count flows instead of bytes
  {"sampling",1,0,0}, // trace has packets sampled with probability 1/sampling
  {"verbose",0,0,0}, // print extra stuff
  {"forceroot",0,0,0}, // forces the root to be in the compressed report
  {"javascript",0,0,0}, // output as javascript variables
  {"timestamps",0,0,0}, // trace contains timestamps 
  {"quality",1,0,0}, // trades off accuracy versus running time
  {"synopsis",1,0,0}, // the synopsis file describing the traffic
  {"seed",1,0,0}, // the seed used by the flow synopsis (we also srand after parsing options)
  {"maxsynsize",1,0,0}, // the maximum size of a synopsis
 {0,0,0,0}
};

char* uninitialized="-";
char* trace=uninitialized;
char* flowrecfile=uninitialized;
char* clusters=uninitialized;
char* expected=uninitialized;
char* synopsis=uninitialized;
int resolution=20; /* threshold = totaltraffic / resolution */
int compressed=1;
int normalize=0;
int countpackets=0;
int countflows=0;
uint4 sampling=1;
int verbose=0;
int forceroot=0;
int javascript=0;
int timestamps=0;
int seed=1;
unsigned maxsynsize=10000000;
#define QUICK     0
#define MEDIUM    1
#define ACCURATE  2
#define PERFECT   3
int quality=PERFECT;

void checkoption(int cond,const char* errorstring){
  if(!cond){
    fprintf(stderr,"Incorrect command line: %s\n",errorstring);
    exit(1);
  }
}

void checktraceoptions(){
  checkoption(countpackets==0,
	      "--countpackets only makes sense when using a trace");
  checkoption(countflows==0,
	      "--countflows only makes sense when using a trace or a synopsis");
  checkoption(sampling==1,
	      "--sampling only makes sense when using a trace or a synopsis");
}

ip srcip("Source IP");
ip destip("Dest. IP");
port srcport("SrcP");
port destport("DestP");
proto protocol("Pro");
record flowID(&srcip,&destip,&protocol,&srcport,&destport);
prefix srcpref(srcip);
prefix destpref(destip);
protop prp(protocol);
prange srcpr(srcport);
prange destpr(destport);
cluster cl(&srcpref,&destpref,&prp,&srcpr,&destpr);
buffer b;

void print_cluster_name(){
  b.offset=0;
  if(verbose){
    cl.prettyprintname(b);
    b.print_char('\0');
    printf("%s\n",b.buff);
  }
}

uint8 read_in_the_trace(hash_map& flowmemory){
  uint8 totaltraffic=0;
  hash_map* fmp=&flowmemory;
  if(quality!=PERFECT){
    fmp = new growing_hash(flowID.getbinsize(),8);
  }
  if(flowrecfile!=uninitialized){
    read_flow_records(flowrecfile,flowID,*fmp,
		      (countpackets?AGGREGATION_COUNT:
		       (countflows?AGGREGATION_CREATE:AGGREGATION_SUM)));
  } else {
    if(synopsis!=uninitialized){
      if(countflows){
	flowsyn fs(&flowID,maxsynsize,seed);
	fs.read_in_syn(synopsis);
	hash_map_iterator* i=fs.build_iterator();
	for(;!i->at_end();++(*i)){
	  *(uint8*)((*fmp)[i->key()])=fs.compensation_factor();
	}
	delete i;
      } else {
	countingsyn cs(&flowID,maxsynsize);
	cs.read_in_syn(synopsis);
	hash_map_iterator* i=cs.build_iterator();
	for(;!i->at_end();++(*i)){
	  *(uint8*)((*fmp)[i->key()])=*(uint8*)i->value()+cs.get_th()-1;
	}
	delete i;
      }
    } else {
      if(timestamps) {
	read_timestamped(trace,flowID,*fmp,
			 (countpackets?AGGREGATION_COUNT:
			  (countflows?AGGREGATION_CREATE:AGGREGATION_SUM)));
      } else {
	read_plain(trace,flowID,*fmp,
		   (countpackets?AGGREGATION_COUNT:
		    (countflows?AGGREGATION_CREATE:AGGREGATION_SUM)));
      }
    }
  }
  hash_map_iterator* i=fmp->build_iterator();
  for(;!i->at_end();++(*i)){
    if(!countflows){
      *((uint8*)i->value())=sampling*(*((uint8*)i->value()));
    }
    totaltraffic+=*((uint8*)i->value());
  }
  delete i;
  if(quality!=PERFECT){
    uint8 threshold=totaltraffic/resolution;
    uint8 z=(quality==QUICK?threshold/10:
	     (quality==MEDIUM?threshold/100:threshold/1000));
    i=fmp->build_iterator();
    for(;!i->at_end();++(*i)){
      if(*((uint8*)i->value())<z){
	if(z*((double)rand())/(RAND_MAX+1.0)<*((uint8*)i->value())){
	  *((uint8*)i->value())=z;
	} else {
	  continue;
	}
      }
      *(uint8*)(flowmemory[i->key()])=*((uint8*)i->value());
    }
    delete i;
    delete fmp;
  }
  return totaltraffic;
}

uint8 totaltraffic;
void print_unidim_report(pattern* p, char* heading, unsigned index,
			 clustergraph* cg){
  clgnlist* report;
  uint8 traffic;
  uint8 threshold=totaltraffic/resolution;
  if(threshold==0){
    threshold=1;
  }
  if(javascript){
    printf("var %s=[",heading);
  } else {
    printf(heading);
  }
  report=build_unidim_report(cg,index,threshold);
  int isfirst=1;
  for(clgnlist* crt=report;crt!=NULL;crt=crt->next){
    if(p->isstar((byte*)(crt->node)-cl.getbinsize()+cl.getoffset(index))){
      continue;
    }
    b.offset=0;
    if(javascript){
      p->print(b,(byte*)(crt->node)-cl.getbinsize()+cl.getoffset(index));
    } else {
      p->prettyprint(b,(byte*)(crt->node)-cl.getbinsize()+cl.getoffset(index));
    }
    b.print_char('\0');
    traffic=cg->counter_i(crt->node,0);
    if(javascript){
      if(isfirst){
	printf("[\"%s\",%2.3f,\"%llu\"]",b.buff,
	       (100.0*traffic)/totaltraffic,traffic);
	isfirst=0;
      } else {
	printf(",\n[\"%s\",%2.3f,\"%llu\"]",b.buff,
	       (100.0*traffic)/totaltraffic,traffic);
      }
    } else {
      printf(" %s %2.3f%%\t%llu\n",b.buff,(100.0*traffic)/totaltraffic,
	     traffic);
    }
  }
  if(javascript){
    printf("];\n");
  }
  freelist(report);
}

int main(int argc,char* argv[]){
  int i;
  /*options stuff*/
  int option_index=0;
  int cc;
  /* parsing the command line options*/
  while((cc=getopt_long_only(argc,argv,"",long_options,&option_index))!=-1){
    if (cc!=0){
      fprintf (stderr,"Problem with options %c .\n",cc);
      exit(1);
    }
    switch (option_index) {
    case 0:
      trace=optarg;
      break;
    case 1:
      flowrecfile=optarg;
      break;
    case 2:
      clusters=optarg;
      break;
    case 3:
      expected=optarg;
      break;
    case 4:
      resolution=atoi(optarg);
      checkoption(resolution>1,"resolution must be above 1");
      checkoption(resolution<=10000,"resolution cannot be above 10000");
      break;
    case 5:
      compressed=0;
      break;
    case 6:
      normalize=1;
      break;
    case 7:
      countpackets=1;
      break;
    case 8:
      countflows=1;
      break;
    case 9:
      sampling=atoi(optarg);
      checkoption(sampling>0,"the sampling rate must be positive");
      break;
    case 10:
      verbose=1;
      break;
    case 11:
      forceroot=1;
      break;
    case 12:
      javascript=1;
      break;
    case 13:
      timestamps=1;
      break;
    case 14:
      switch(optarg[0]){
      case 'q':
	quality=QUICK;
	break;
      case 'm':
	quality=MEDIUM;
	break;
      case 'a':
	quality=ACCURATE;
	break;
      case 'p':
	quality=PERFECT;
	break;
      default:
	checkoption(0,"quality must be quick, medium, accurate or perfect");
      }
      break;
    case 15:
      synopsis=optarg;
      break;
    case 16:
      seed=atoi(optarg);
      break;
    case 17:
      maxsynsize=atoi(optarg);
      checkoption(maxsynsize>=1000,"maxsynsize must be at least 1000");
      break;
    default:
      assume(0,"Major command line argument parsing error.\n");
    }
  }

  if(normalize || expected!= uninitialized){
    checkoption(strcmp(expected,"-")!=0,"--expected must specify a file");
    checkoption(trace==uninitialized,
		"cannot use --trace when computing deltas");
    checkoption(flowrecfile==uninitialized,
		"cannot use --flowrecfile when computing deltas");
    checkoption(synopsis==uninitialized,
		"cannot use --synopsis when computing deltas");
    checkoption(forceroot==0,
		"cannot use --forceroot when computing deltas");
    checkoption(javascript==0,
		"javascript not supported yet when computing deltas");
    checktraceoptions();
    clustergraph cg(cl,2);
    cg.read_in_clusters(clusters,0);
    cg.read_in_clusters(expected,1);
    if(normalize){
      normalizeclusterreports(&cg);
    }
    totaltraffic=(cg.totaltraffic(0)>cg.totaltraffic(1)?
		  cg.totaltraffic(0):cg.totaltraffic(1));
    uint8 threshold=totaltraffic/resolution;
    if(threshold==0){
      threshold=1;
    }
    clgnlist* report=build_delta_report(&cg,threshold);
    sint8 traffic;
    print_cluster_name();
    for(clgnlist* crt=report;crt!=NULL;crt=crt->next){
      b.offset=0;
      cl.prettyprint(b,(byte*)(crt->node)-cl.getbinsize());
      b.print_char('\0');
      traffic=cg.counter_i(crt->node,0)-cg.counter_i(crt->node,1);
      printf(" %s %lld\n",b.buff,traffic);
    }
    freelist(report);
    return 0;
  }
  checkoption(clusters==uninitialized||trace==uninitialized,
	      "cannot use both --trace and --clusters");
  checkoption(clusters==uninitialized||flowrecfile==uninitialized,
	      "cannot use both --flowrecfile and --clusters");
  checkoption(flowrecfile==uninitialized||trace==uninitialized,
	      "cannot use both --trace and --flowrecfile");
  checkoption(synopsis==uninitialized||trace==uninitialized,
	      "cannot use both --trace and --synopsis");
  checkoption(clusters==uninitialized||synopsis==uninitialized,
	      "cannot use both --synopsis and --clusters");
  checkoption(synopsis==uninitialized||trace==uninitialized,
	      "cannot use both --trace and --synopsis");
  checkoption((!countflows)||(!countpackets),
	      "cannot use both --countflows and --countpackets");
  if(!compressed){/* The user wants an uncompressed report */
    checkoption(clusters==uninitialized,
		"cannot use --clusters with --uncompressed");
    checkoption(forceroot==0,
		"cannot use --forceroot with --uncompressed");
    checkoption(javascript==0,
		"javascript not supported for uncompressed reports");
    growing_hash flowmemory(flowID.getbinsize(),8);
    uint8 totaltraffic=read_in_the_trace(flowmemory);
    growing_hash hvclusters(flowID.getcluster().getbinsize(),8);
    uint8 threshold=totaltraffic/resolution;
    if(threshold==0){
      threshold=1;
    }
    find_high_volume_clusters(flowID,flowmemory,threshold,hvclusters);
    print_cluster_name();
    for(growing_hash_iterator i(hvclusters);!i.at_end();++i){
      b.offset=0;
      cl.prettyprint(b,i.key());
      b.print_char('\0');
      printf("%s   %llu\n",b.buff,*((uint8*)i.value()));
    }
    return 0;
  }
  /* We produce a compressed report */
  clustergraph cg(cl,3);
  if(clusters!=uninitialized){ /* Initialize based on report */
    checktraceoptions();
    cg.read_in_clusters(clusters);
    if(cg.hvclmem.size()!=0){
      totaltraffic=cg.totaltraffic();
    } else {
      totaltraffic=0;
    }
  } else { /* Read in trace and compute large clusters. */
    growing_hash flowmemory(flowID.getbinsize(),8);
    totaltraffic=read_in_the_trace(flowmemory);
    uint8 threshold=totaltraffic/resolution;
    if(threshold==0){
      threshold=1;
    }
    cg.find_hvclusters(flowID,flowmemory,threshold);
  }
  uint8 threshold=totaltraffic/resolution;
  if(threshold==0){
    threshold=1;
  }
  if(javascript){
    printf("var traffic=\"%llu\";\nvar threshold=\"%llu\";\n",totaltraffic,
	   threshold);
    print_unidim_report(&prp,"proto",2,&cg);
    print_unidim_report(&srcpref,"source_ip",0,&cg);
    print_unidim_report(&destpref,"dest_ip",1,&cg);
    print_unidim_report(&srcpr,"source_port",3,&cg);
    print_unidim_report(&destpr,"dest_port",4,&cg);
    printf("var combined=[");
  } else {
    if(verbose){
      printf("Total traffic %llu, threshold %llu\n",totaltraffic,threshold);
      print_unidim_report(&prp,"Protocol breakdown\n",2,&cg);
      print_unidim_report(&srcpref,"Source IP breakdown\n",0,&cg);
      print_unidim_report(&destpref,"Destination IP breakdown\n",1,&cg);
      print_unidim_report(&srcpr,"Source port breakdown\n",3,&cg);
      print_unidim_report(&destpr,"Destination port breakdown\n",4,&cg);
      printf("Compressed multidimensional report\n");
      print_cluster_name();
    }
  }
  clgnlist* report=build_multidim_report(&cg,threshold,forceroot);
  uint8 traffic;
  int isfirst=1;
  for(clgnlist* crt=report;crt!=NULL;crt=crt->next){
    b.offset=0;
    cl.prettyprint(b,(byte*)(crt->node)-cl.getbinsize());
    b.print_char('\0');
    traffic=cg.counter_i(crt->node,0);
    double ratio= 100.0*traffic/cg.doublecounter_i(crt->node,2);
    if(javascript){
      if(isfirst){
	printf("[");
	isfirst=0;
      } else {
	printf(",\n[");
      }
      for(i=0;i<cl.getnumthings()-1;i++){
	b.offset=0;
	cl.getpattern(i).print(b,(byte*)(crt->node)-cl.getbinsize()+
			       cl.getoffset(i));
	b.print_char('\0');
	printf("\"%s\",",b.buff);
      }
      b.offset=0;
      cl.getpattern(i).print(b,(byte*)(crt->node)-cl.getbinsize()+
			     cl.getoffset(i));
      b.print_char('\0');
      printf("\"%s\",\"%llu\",%2.1f]",b.buff,traffic,ratio);
    } else {
      printf(" %s %llu    %2.1f%%\n",b.buff,traffic,ratio);
    }
  }
  freelist(report);
  if(javascript){
    printf("];\n");
  }
}
