/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 10/09/2003 */

/* The netflowsplitter splits the NetFlow data by time and
   category. The categories are defined by the categories.txt file
   that netflowsplitter reads in from the current directory. For its
   syntax see splitter. One separate flow record file (without the
   timestamps) is created for each category for each one minute time
   bin. Each bin has its own directory named after the unix time it
   starts at. Because a NetFlow record can span more than one bin, we
   work with many bins simultaneously. The activetimeout command line
   argument (measured in seconds) tells us how far behind the most
   recent time the first packet of a subsequent flow record can
   fall. When an old bin "expires", the flow record files in it are
   closed and we execute an external program that is supposed to
   process them. */

#include <stdio.h>
#include <glob.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h> //detecting future flows
#include "../datatypes/ip/ip.h"
#include "../datatypes/port/port.h"
#include "../datatypes/protocol/proto.h"
#include "../datatypes/cluster.h"
#include "../datatypes/counter64.h"
#include "../datatypes/timestamp.h"
#include "../readwrite/basicfilereader.h"
#include "../readwrite/flowexportreader.h"
#include "../readwrite/assume.h"

ip srcip("Source IP");
ip destip("Dest. IP");
port srcport("SrcP");
port destport("DestP");
proto protocol("Pro");
counter64 octets("Bytes");
counter64 packets("Packets");
timestamp first("First");
timestamp last("Last");
compound flowrec(&first,&last,&srcip,&destip,&protocol,&srcport,&destport,&octets,&packets);
compound flowID(&srcip,&destip,&protocol,&srcport,&destport);
prefix srcpref(srcip);
prefix destpref(destip);
protop prp(protocol);
prange srcpr(srcport);
prange destpr(destport);
cluster cl(&srcpref,&destpref,&prp,&srcpr,&destpr);

int activetimeout;

#define MAXCATS 50
int numcats=0;
char* catnames[MAXCATS];

#define MAXACTIVEBINS 200
#define ILENGTH 60

int numbins; /* The number of bins active at the same time. */

uint8 startofint[MAXACTIVEBINS+1];
FILE* frfiles[MAXACTIVEBINS][MAXCATS];

#define MAXRULES 200
int numrules=0;
typedef struct ruleinfo{
  byte clusterbytes[20];
  int cat;
} ruleinfo;
ruleinfo rules[MAXRULES];

byte *linebytes; /* The current packet parsed */
char* programname;

buffer flowIDtext;

void read_in_category_file(){
  basicfilereader catfile("categories.txt");
  const char* categoryname;
  int i;
  while(catfile.getnextline()){
    if(!(numrules<MAXRULES)){
      fprintf(stderr,"Can have at most %d clusters in categories.txt\n",
	      MAXRULES);
      exit(103);
    }
    if(!cl.parse(catfile.b,rules[numrules].clusterbytes)){
      catfile.printbuffererror();
    }
    catfile.b.skipspaces();
    if(!isalpha(catfile.b.buff[catfile.b.offset])){
      catfile.printbuffererror("Category name must start with alphabetic character");
    }
    categoryname=catfile.b.buff+catfile.b.offset;
    while(isalnum(catfile.b.buff[catfile.b.offset]) || 
          catfile.b.buff[catfile.b.offset]=='_'){
      catfile.b.offset++;
    }
    catfile.b.buff[catfile.b.offset]='\0';
    if(strlen(categoryname)>19){
      fprintf(stderr,"Category name %s exceeds 19 characters\n",categoryname);
      exit(109);
    }
    for(i=0; i<numcats; i++){
      if(strcmp(categoryname,catnames[i])==0){
	rules[numrules].cat=i;
	numrules++;
	break;
      }
    }
    if(i<numcats){
      continue;
    }
    /* This line defines a new category */
    rules[numrules].cat=numcats;
    numrules++;
    if(!(numcats<MAXCATS)){
      fprintf(stderr,"Can have at most %d categories in categories.txt",
	      MAXCATS);
      exit(104);
    }
    catnames[numcats]=(char*)malloc(strlen(categoryname)+1);
    assumenotnull(catnames[numcats]);
    strcpy(catnames[numcats],categoryname);
    numcats++;
  }
}

/* Reports to higher level software that the splitter is done with a
   bin. */
void report_bin(uint8 binstartsec){
  /* Moving the directory with the flow record files. */
  char olddir[30];
  char newdir[40];
  sprintf(olddir,"tmp/%lld",binstartsec);
  struct stat junk;
  if(stat(olddir,&junk)==-1){
    return;
  }
  sprintf(newdir,"split_raw_data/%lld",binstartsec);
  assume2(rename(olddir,newdir)==0,"Trouble renaming %s to %s\n",
	  olddir,newdir);
  char intervalend[21];
  sprintf(intervalend,"%llu",binstartsec+ILENGTH);
  pid_t childpid;
  childpid=fork();
  assume(childpid!=-1,"Error creating shild process\n");
  if(childpid==0){ /* Child process, we run the reporting program */
    execlp(programname,programname,intervalend,NULL);
    fprintf(stderr,"exec failed for %s %s\n",programname,intervalend);
    perror(NULL);
    exit(217);
  } else { /* Parent process, we wait for the child to finish */
    int status;
    waitpid(childpid,&status,0);
    if(!WIFEXITED(status)){
      if(WIFSIGNALED(status)){
	fprintf(stderr,"Child %s %s killed by signal %d\n",
		programname,intervalend,WTERMSIG(status));
	exit(105);
      }
      fprintf(stderr,"Something went wrong with child %s %s\n",
	      programname,intervalend);
      exit(106);
    }
    if(WEXITSTATUS(status)!=0){
      fprintf(stderr,"Child %s %s exited with error code %d\n",
	      programname,intervalend,WEXITSTATUS(status));
      exit(107);
    }
  }
}

/* Initializes the globals that hold information about the bins and
   cleans up (and reports) old bins left by previous runs. */

void initialization(uint8 binstartsec){
  uint8 firstbinstart=binstartsec-numbins*ILENGTH+ILENGTH;
  glob_t globbuf;
  if(glob("tmp/*",GLOB_ERR|GLOB_NOSORT,NULL,&globbuf)==0){
    /* Found some old stuff, reporting it. */
    uint8* binstoreport=(uint8*)malloc(globbuf.gl_pathc*8);
    assumenotnull(binstoreport);
    int nbinstr=0;
    for(int i=0;i<globbuf.gl_pathc;i++){
      uint8 crtbin=0;
      int k;
      for(k=4;globbuf.gl_pathv[i][k]<='9' && globbuf.gl_pathv[i][k]>='0';
	  k++){
	crtbin*=10;
	crtbin+=globbuf.gl_pathv[i][k]-'0';
      }
      if(globbuf.gl_pathv[i][k]=='\0' && k>10 && crtbin<firstbinstart){ 
	/* Looks like an old expired bin */
	int cb;
	for(cb=nbinstr-1;cb>=0;cb--){
	  if(binstoreport[cb]>crtbin){
	    binstoreport[cb+1]=binstoreport[cb];
	  } else {
	    break;
	  }
	}
	binstoreport[cb+1]=crtbin;
	nbinstr++;
      }
    }
    for(int i=0;i<nbinstr;i++){
      report_bin(binstoreport[i]);
    }
  }
  globfree(&globbuf);
  for(int b=0;b<numbins;b++){
    startofint[b]=firstbinstart+b*ILENGTH;
    for(int c=0;c<numcats;c++){
      frfiles[b][c]=NULL;
    }
  }
  startofint[numbins]=binstartsec+ILENGTH;
}

/* Shifts the bins to make space for a new one and reports the oldest
   one. */
void shift_bins(){
  if(frfiles[0][0]!=NULL){
    for(int c=0;c<numcats;c++){
     fclose(frfiles[0][c]);
    }
  }
  report_bin(startofint[0]);
  for(int b=0;b<numbins-1;b++){
    startofint[b]=startofint[b+1];
    for(int c=0;c<numcats;c++){
      frfiles[b][c]=frfiles[b+1][c];
    }
  }
  startofint[numbins-1]=startofint[numbins];
  startofint[numbins]=startofint[numbins]+ILENGTH;
  for(int c=0;c<numcats;c++){
    frfiles[numbins-1][c]=NULL;
  }
}

/* Adds to bin "binstartsec", file corresponding to category cat, a
   number of bytes and packets of the current flow.*/
void add_to_bin(uint8 binstartsec, unsigned cat, uint8 bytes, uint8 packets){
  if(binstartsec<startofint[0]){
    fprintf(stderr,"Data exceeds the activetimeout parameter.\n");
    flowrec.debugprint(linebytes);
    fprintf(stderr,"Most recent bin used is %lld, number of bins is %d\n",
	    startofint[numbins-1],numbins);
    exit(153);
  }
  if(startofint[0]==0){
    initialization(binstartsec);
  }
  while(binstartsec>startofint[numbins-1]){
    shift_bins();
  }
  int binindex=(binstartsec-startofint[0])/ILENGTH;
  if(frfiles[binindex][0]==NULL){ /* Must open the files in the bin. */
    char outputfilename[BUFFERSIZE];
    sprintf(outputfilename,"tmp/%lld",startofint[binindex]);
    mkdir(outputfilename,0x1ED);
    for(int i=0;i<numcats;i++){
      if (strcmp(catnames[i] , "discard") == 0){
        sprintf(outputfilename, "/dev/null");
      } else {
        sprintf(outputfilename,"tmp/%lld/%s.frt",startofint[binindex],
		catnames[i]);
      }
      if((frfiles[binindex][i]=fopen(outputfilename,"a"))==NULL){
	printf("Can not open %s for writing.\n",outputfilename);
	exit(102);
      }
    }
  }
  fprintf(frfiles[binindex][cat],"%s %lld %lld\n",flowIDtext.buff,bytes,packets);
}

void close_files(){
  for(int b=0;b<numbins-1;b++){
    if(frfiles[b][0]!=NULL){
      for(int c=0;c<numcats;c++){
	fclose(frfiles[b][c]);
      }
    }
  }
}

int main(int argc,char* argv[]){
  int mycat,i;
  struct timeval myt;
  if(argc!=3){
    fprintf(stderr,"Usage netflowsplitter activetimeout reporting_program_name\n");
    exit(108);
  } else {
    activetimeout=atoi(argv[1]);
    numbins=2+activetimeout/ILENGTH;
    assume1(numbins>0 && numbins<=MAXACTIVEBINS,
	   "The first command line argument must be between 1 and %d\n",
	   ILENGTH*MAXACTIVEBINS);
    programname=argv[2];
  }
  assume(cl.getbinsize()==20,"Internal error, cluster too big\n");
  read_in_category_file();
  if (gettimeofday(&myt, NULL)) {
    fprintf(stderr, "Could not gettimeofday!\n");
    exit(101);
  }
 
  startofint[0]=0; // Marks globals as uninitialized
  flowexportreader input("-",0);
  while(linebytes=input.getnextrecord()){
    for(i=0;i<numrules;i++){
      if(cl.matches(rules[i].clusterbytes,linebytes+16)){
	mycat=rules[i].cat;
	break;
      }
    }
    if(i==numrules){
       fprintf(stderr,"FlowID in input does not match any category.\n");
       flowID.debugprint(linebytes+16);
       exit(101);
    }
    flowIDtext.offset=0;
    flowID.print(flowIDtext,linebytes+16);
    flowIDtext.print_char('\0');
    uint8 first=*(uint8*)linebytes;
    uint8 last=*(uint8*)(linebytes+8);
    uint8 bytes=*(uint8*)(linebytes+31);
    uint8 packets=*(uint8*)(linebytes+39);
    first=first/1000000000;// Converting to seconds
    last=last/1000000000;// Converting to seconds

    if (first > myt.tv_sec) {//flow is in the future
      continue;  //skip it
    }
  
    if(first==last){
      add_to_bin(first-first%ILENGTH,mycat,bytes,packets);
      continue;
    }
    last=last+1; //Rounding up
    uint8 flowlen=last-first;
    uint8 packetssofar=0;
    uint8 bytessofar=0;
    for(uint8 i=first-first%ILENGTH;i<last;i+=ILENGTH){
      uint8 endofbin=((i+ILENGTH)<last?(i+ILENGTH):last);
      uint8 packetstoendofbin=((endofbin-first)*(packets-1))/flowlen+1;
      uint8 bytestoendofbin=(packetstoendofbin*bytes)/packets;
      add_to_bin(i,mycat,bytestoendofbin-bytessofar,
		 packetstoendofbin-packetssofar);
      packetssofar=packetstoendofbin;
      bytessofar=bytestoendofbin;
    }
  }
  close_files();
}
