/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 07/23/2003 */

/* The splitter splits the input trace by time and category. The
   categories are defined by the categories.txt file that the splitter
   reads in from the current directory. Each line contains a cluster
   and a category name (alphabetic character followed by _s and
   alphanumerics up to 19 characters long). Each packet is mapped to
   the category of the first cluster it matches. There can be more
   than one cluster mapping to the same category. The first line where
   a category appears defines the color associated with it and a
   longer textual description. One separate trace file (without the
   timestamps) is created for each category. After each minute of the
   trace, the split traces are closed, we execute an external program
   that is supposed to process them and we start anew. The splitter
   also "fills in the gaps" by reporting empty results for the minutes
   that have been skipped due to gaps in the trace or aggressive
   sampling. The activetimeout command line parameter tells the
   splitter how far back to look when filling gaps. */

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "../datatypes/ip/ip.h"
#include "../datatypes/port/port.h"
#include "../datatypes/protocol/proto.h"
#include "../datatypes/cluster.h"
#include "../datatypes/counter64.h"
#include "../datatypes/timestamp.h"
#include "../readwrite/basicfilereader.h"
#include "../readwrite/asciifile.h"
#include "../readwrite/assume.h"

ip srcip("Source IP");
ip destip("Dest. IP");
port srcport("SrcP");
port destport("DestP");
proto protocol("Pro");
counter64 size("Size");
timestamp ts("TimeStamp");
compound flowID(&srcip,&destip,&protocol,&srcport,&destport);
compound flowID_and_size(&srcip,&destip,&protocol,&srcport,&destport,&size);
prefix srcpref(srcip);
prefix destpref(destip);
protop prp(protocol);
prange srcpr(srcport);
prange destpr(destport);
cluster cl(&srcpref,&destpref,&prp,&srcpr,&destpr);
int lookback; // How many intervals back to look to "fill in the gaps"

#define MAXCATS 50
int numcats=0;
typedef struct catinfo{
  char* name;
  FILE* F;
} catinfo;
catinfo cats[MAXCATS];

#define MAXRULES 200
int numrules=0;
typedef struct ruleinfo{
  byte clusterbytes[20];
  int cat;
} ruleinfo;
ruleinfo rules[MAXRULES];

void read_in_category_file(){
  basicfilereader catfile("categories.txt");
  const char* categoryname;
  int i;
  while(catfile.getnextline()){
    if(!(numrules<MAXRULES)){
      fprintf(stderr,"Can have at most %d clusters in categories.txt\n",
	      MAXRULES);
      exit(103);
    }
    if(!cl.parse(catfile.b,rules[numrules].clusterbytes)){
      catfile.printbuffererror();
    }
    catfile.b.skipspaces();
    if(!isalpha(catfile.b.buff[catfile.b.offset])){
      catfile.printbuffererror("Category name must start with alphabetic character");
    }
    categoryname=catfile.b.buff+catfile.b.offset;
    while(isalnum(catfile.b.buff[catfile.b.offset]) || 
          catfile.b.buff[catfile.b.offset]=='_'){
      catfile.b.offset++;
    }
    catfile.b.buff[catfile.b.offset]='\0';
    if(strlen(categoryname)>19){
      fprintf(stderr,"Category name %s exceeds 19 characters\n",categoryname);
      exit(109);
    }
    for(i=0; i<numcats; i++){
      if(strcmp(categoryname,cats[i].name)==0){
	rules[numrules].cat=i;
	numrules++;
	break;
      }
    }
    if(i<numcats){
      continue;
    }
    /* This line defines a new category */
    rules[numrules].cat=numcats;
    numrules++;
    if(!(numcats<MAXCATS)){
      fprintf(stderr,"Can have at most %d categories in categories.txt",
	      MAXCATS);
      exit(104);
    }
    cats[numcats].name=(char*)malloc(strlen(categoryname)+1);
    assumenotnull(cats[numcats].name);
    strcpy(cats[numcats].name,categoryname);
    numcats++;
  }
}

#define ILENGTH 60

/* The timestamp with the end of the current interval. */
byte endofint[8]={'\0','\0','\0','\0','\0','\0','\0','\0'};
byte *linebytes; /* The current packet parsed */
char* programname;

/* Reports that the splitter is done generating data for the interval
   starting at startofint. */
void report_end_of_interval(uint8 startofint){
  /* Moving the directory with the traces. */
  char olddir[BUFFERSIZE];
  char newdir[BUFFERSIZE];
  sprintf(newdir,"split_raw_data/%lld",startofint);
  sprintf(olddir,"tmp/%lld",startofint);
  rename(olddir,newdir);
  char intervalend[21];
  sprintf(intervalend,"%llu",startofint+ILENGTH);
  pid_t childpid;
  childpid=fork();
  assume(childpid!=-1,"Error creating child process!\n");
  if(childpid==0){ /* Child process, we run the reporting program */
    execlp(programname,programname,intervalend,NULL);
    fprintf(stderr,"exec failed for %s %s\n",programname,intervalend);
    perror(NULL);
    exit(217);
  } else { /* Parent process, we wait for the child to finish */
    int status;
    waitpid(childpid,&status,0);
    if(!WIFEXITED(status)){
      if(WIFSIGNALED(status)){
	fprintf(stderr,"Child %s %s killed by signal %d\n",
		programname,intervalend,WTERMSIG(status));
	exit(105);
      }
      fprintf(stderr,"Something went wrong with child %s %s\n",
	      programname,intervalend);
      exit(106);
    }
    if(WEXITSTATUS(status)!=0){
      fprintf(stderr,"Child %s %s exited with error code %d\n",
	      programname,intervalend,WEXITSTATUS(status));
      exit(107);
    }
  }
}


void move_to_next_interval(){
  char dir_name[BUFFERSIZE];
  char outputfilename[BUFFERSIZE+5];
  struct stat junk;
  uint8 startofint;
  int i;
  if(*((uint8*)endofint)==0){ /* First interval */
    ts.setseconds(endofint,((ts.getseconds(linebytes)/ILENGTH+1)*ILENGTH));
  } else {
    for(i=0;i<numcats;i++){
      fclose(cats[i].F);
    }
    ts.setseconds(endofint,ts.getseconds(endofint)+ILENGTH);
  }
  startofint=ts.getseconds(endofint)-ILENGTH;
  for(i=lookback;i>0;i--){
    sprintf(dir_name,"tmp/%lld",startofint-i*ILENGTH);
    if(stat(dir_name,&junk)!=-1){
      break;
    }
  }
  for(;i>0;i--){
    report_end_of_interval(startofint-i*ILENGTH);
  }
  sprintf(dir_name,"tmp/%lld",startofint);
  mkdir(dir_name,0x1ED);
  for(int i=0;i<numcats;i++){
    if (strcmp(cats[i].name,"discard")==0) {
      sprintf(outputfilename, "/dev/null");
    } else {
      sprintf(outputfilename,"%s/%s.ntt",dir_name,cats[i].name);
    }
    if((cats[i].F=fopen(outputfilename,"a"))==NULL){
      printf("Can not open %s for writing.\n",outputfilename);
      exit(102);
    }
  }
}

int main(int argc,char* argv[]){
  buffer output;
  int mycat,i;
  compound traceline(&ts,&srcip,&destip,&protocol,&srcport,&destport,&size);
  if(argc!=3){
    fprintf(stderr,"Usage splitter activetimeout reporting_program_name\n");
    exit(108);
  } else {
    lookback=atoi(argv[1])/ILENGTH;
    programname=argv[2];
  }
  assume(cl.getbinsize()==20,"Internal error, cluster too big\n");
  assume(ts.getbinsize()==8,"Internal error, timestamps wrong size\n");
  read_in_category_file();
  asciifilereader trace("-",traceline,0);
  while(linebytes=trace.getnextrecord()){
    while(!ts.isbefore(linebytes,endofint)){
      move_to_next_interval();
    }
    for(i=0;i<numrules;i++){
      if(cl.matches(rules[i].clusterbytes,linebytes+8)){
	mycat=rules[i].cat;
	break;
      }
    }
    if(i==numrules){
      fprintf(stderr,"FlowID in input does not match any category.\n");
      flowID.debugprint(linebytes+8);
      exit(101);
    }
    output.offset=0;
    flowID_and_size.print(output,linebytes+8);
    output.print_char('\0');
    fprintf(cats[mycat].F,"%s\n",output.buff);
  }
  for(int i=0;i<numcats;i++){
    fclose(cats[i].F);
  }
  //report_end_of_interval();
}
