/* See AutoFocus/doc/COPYING for copyright/licensing information. */
/* cestan@cs.ucsd.edu 10/13/2003 */

/* The summarizer takes in packet header traces, flow record data or
   synopses and produces traffic synopses (summaries). */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "../getopt/getopt.h"
#include "../readwrite/asciifile.h"
#include "../readwrite/assume.h"
#include "../datatypes/ip/ip.h"
#include "../datatypes/port/port.h"
#include "../datatypes/protocol/proto.h"
#include "../datatypes/compound.h"
#include "../datatypes/counter64.h"
#include "../algorithms/countingsynopsis.h"
#include "../algorithms/flowsynopsis.h"

static struct option long_options[] =
{
  {"trace",1,0,0}, // the trace
  {"flowrecfile",1,0,0}, // the input file with flow records
  {"synopses",1,0,0}, // the synopsis files in the input (comma separated)
  {"countpackets",0,0,0}, // count packets instead of bytes
  {"countflows",0,0,0}, // count flows instead of bytes
  {"seed",1,0,0}, // the seed used by flow synopses
  {"maxsize",1,0,0}, // the maximum size of the synopsis
  {0,0,0,0}
};

char* uninitialized="-";
char* trace=uninitialized;
char* flowrecfile=uninitialized;
char* synopses=uninitialized;
int countpackets=0;
int countflows=0;
unsigned seed=1;
unsigned maxsize=1000000;

void checkoption(int cond,const char* errorstring){
  if(!cond){
    fprintf(stderr,"Incorrect command line: %s\n",errorstring);
    exit(1);
  }
}

ip srcip("Source IP");
ip destip("Dest. IP");
port srcport("SrcP");
port destport("DestP");
proto protocol("Pro");
compound flowID(&srcip,&destip,&protocol,&srcport,&destport);
buffer b;

int main(int argc,char* argv[]){
  int i;
  /*options stuff*/
  int option_index=0;
  int cc;
  /* parsing the command line options*/
  while((cc=getopt_long_only(argc,argv,"",long_options,&option_index))!=-1){
    if (cc!=0){
      fprintf (stderr,"Problem with options %c .\n",cc);
      exit(1);
    }
    switch (option_index) {
    case 0:
      trace=optarg;
      break;
    case 1:
      flowrecfile=optarg;
      break;
    case 2:
      synopses=optarg;
      break;
    case 3:
      countpackets=1;
      break;
    case 4:
      countflows=1;
      break;
    case 5:
      seed=atoi(optarg);
      break;
    case 6:
      maxsize=atoi(optarg);
      break;
    default:
      assume(0,"Major command line argument parsing error.\n");
    }
  }
  srand(seed);

  checkoption(flowrecfile==uninitialized||trace==uninitialized,
	      "cannot use both --trace and --flowrecfile");
  checkoption((!countflows)||(!countpackets),
	      "cannot use both --countflows and --countpackets");

  countingsyn csyn(&flowID,maxsize);
  flowsyn fsyn(&flowID,maxsize,seed);
  /* Read in the synopses */
  if(synopses!=uninitialized){
    char* endofsynopses=synopses+strlen(synopses);
    char* crt;
    for(crt=synopses;crt[0]!=','&&crt[0]!='\0';crt++){}
    crt[0]='\0';
    if(countflows){
      fsyn.read_in_syn(synopses);
    } else {
      csyn.read_in_syn(synopses);
    }
    synopses=crt+1;
    while(synopses<endofsynopses){
      for(crt=synopses;crt[0]!=','&&crt[0]!='\0';crt++){}
      crt[0]='\0';
      if(countflows){
	fsyn.add_syn(synopses);
      } else {
	csyn.add_syn(synopses);
      }
      synopses=crt+1;
    }
  }
  byte* crtid;
  if(trace!=uninitialized){
    /* Read in the trace */
    counter64 c("Packet size");
    compound linestructure(&flowID,&c);
    asciifilereader f(trace,linestructure);
    if(countflows){
      while(crtid=f.getnextrecord()){
	fsyn.addpacket(crtid);
      }
    } else {
      if(countpackets){
	while(crtid=f.getnextrecord()){
	  csyn.addpacket(crtid);
	}
      } else {
	unsigned idlen=flowID.getbinsize();
	while(crtid=f.getnextrecord()){
	  csyn.addpacket(crtid,*(uint8*)(crtid+idlen));
	}
      }
    }
  }
  if((trace==uninitialized && synopses==uninitialized) 
     || flowrecfile!=uninitialized){
    /* Read in the flow records */
    counter64 bytes("Bytes");
    counter64 packets("Packets");
    compound linestructure(&flowID,&bytes,&packets);
    asciifilereader f(flowrecfile,linestructure);
    if(countflows){
      while(crtid=f.getnextrecord()){
	fsyn.addpacket(crtid);
      }
    } else {
      unsigned offset;
      if(countpackets){
	offset=flowID.getbinsize()+8;
      } else {
	offset=flowID.getbinsize();
      }
      while(crtid=f.getnextrecord()){
	csyn.addpacket(crtid,*(uint8*)(crtid+offset));
      }
    }
  }
  /* write out the result */
  if(countflows){
    fsyn.write_out_syn("-");
  } else {
    csyn.write_out_syn("-");
  }
}
